#include <time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#if defined(HAVE_CLAMAV)
# include <clamav.h>
# ifndef CLAMAV_H
#  define CLAMAV_H
# endif
#endif

#include <glib.h>
#include "../include/disk.h"

#include "avscanop.h"
#include "config.h"


/* Utilities */
static gchar *FORMATED_NUMBER_STRING(gulong i);
static gchar *TIME_LAPSED_STRING(gulong dt);
static void PRINT_ERR(const gchar *path, const gchar *mesg);
static void PRINT_SCANNING(const gchar *path);
static void PRINT_SCANNING_DIR(const gchar *path);
static void PRINT_REPORT(
	gulong time_start, gulong time_end,
	gulong infected_files,
	gulong size_scanned_mb,
	gulong files_scanned 
);

static gchar *AVScanCopyShortenString(const gchar *s, gint max);


#if defined(CLAMAV_H)
typedef struct {
	struct cl_node	*database;
	gint		database_entries;
	gulong		files_scanned,
			size_scanned_mb,
			infected_files;
} ClamAVData;
static ClamAVData *AVScanOPClamAVInit(const gchar *db_path);
static void AVScanOPClamAVShutdown(ClamAVData *d);
static gint AVScanOPClamAVScanFile(
	ClamAVData *d, const gchar *path,
	gint *stop_count
);
static gint AVScanOPClamAVScanDir(
	ClamAVData *d, const gchar *path,
	gboolean recursive,
	gboolean executables_only,
	gint *stop_count
);
#endif

/* Front End */
gint AVScanOP(
	const gchar *db_path,
	GList *path_list,
	gboolean recursive,
	gboolean executables_only,
	gint *stop_count
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Returns a statically allocated string describing the formatted
 *	number.
 */
static gchar *FORMATED_NUMBER_STRING(gulong i)
{
	gint comma_cnt, slen;
	gchar ss[80], *ss_ptr;             
	static gchar ts[80], *ts_ptr;


	g_snprintf(ss, sizeof(ss), "%ld", i);
	slen = STRLEN(ss);
			  
	/* 3 digits or less? (no commas needed) */
	if(slen <= 3)
	{
	    strcpy(ts, ss);
	    return(ts);
	}

	ts_ptr = ts;
	ss_ptr = ss;

	/* Initialize comma counter */
	comma_cnt = slen % 3;
	if(comma_cnt <= 0)
	    comma_cnt = 3;

	/* Iterate through size string until end is reached */
	while(*ss_ptr != '\0')
	{
	    /* Reset comma counter and put in a comma? */
	    if(comma_cnt <= 0)
	    {
		*ts_ptr++ = ',';
		comma_cnt = 3;
	    }

	    *ts_ptr++ = *ss_ptr++;
	    comma_cnt--;
	}

	/* Null terminate return string */
	*ts_ptr = '\0';

	return(ts);
}

/*
 *	Returns a statically allocated string describing the time
 *	lapsed specified by dt.
 */
static gchar *TIME_LAPSED_STRING(gulong dt)
{
	gulong ct, ct2;
	static gchar buf[80];

	/* Less than one second? */   
	if(dt < 1)
	{
	    g_snprintf(
		buf, sizeof(buf),
		"less than one second"
	    );
	}
	/* Less than one minute? */
	else if(dt < (1 * 60))
	{
	    ct = MAX(dt / 1, 1);
	    g_snprintf(
		buf, sizeof(buf), 
		"%ld second%s",
		ct, ((ct == 1) ? "" : "s")
	    );
	}
	/* Less than one hour? */
	else if(dt < (60 * 60))
	{
	    ct = MAX(dt / 60, 1);
	    ct2 = MAX(dt / 1, 1) % 60;
	    g_snprintf(
		buf, sizeof(buf),
		"%ld minute%s %ld second%s",
		ct, ((ct == 1) ? "" : "s"),
		ct2, ((ct2 == 1) ? "" : "s")
	    );
	}
	/* Less than one day? */
	else if(dt < (24 * 60 * 60))
	{
	    ct = MAX(dt / 60 / 60, 1);
	    ct2 = MAX(dt / 60, 1) % 60;
	    g_snprintf(
		buf, sizeof(buf),
		"%ld hour%s %ld minute%s",
		ct, ((ct == 1) ? "" : "s"),
		ct2, ((ct2 == 1) ? "" : "s")
	    );
	}
	/* Less than one week? */     
	else if(dt < (7 * 24 * 60 * 60))
	{
	    ct = MAX(dt / 60 / 60 / 24, 1);
	    ct2 = MAX(dt / 60 / 60, 1) % 24;
	    g_snprintf(
		buf, sizeof(buf),
		"%ld day%s %ld hour%s",
		ct, ((ct == 1) ? "" : "s"), 
		ct2, ((ct2 == 1) ? "" : "s")
	    );
	}
	/* Less than one month (30 days)? */
	else if(dt < (30 * 24 * 60 * 60))
	{
	    ct = MAX(dt / 60 / 60 / 24 / 7, 1);
	    ct2 = MAX(dt / 60 / 60 / 24, 1) % 7;
	    g_snprintf(
		buf, sizeof(buf),
		"%ld week%s %ld day%s",
		ct, ((ct == 1) ? "" : "s"),
		ct2, ((ct2 == 1) ? "" : "s")
	    );
	}
	/* Less than 6 months ago? */
#if 0
	else if(dt < (6 * 30 * 24 * 60 * 60))
#else
	else
#endif
	{
	    ct = MAX(dt / 60 / 60 / 24 / 30, 1);
	    ct2 = MAX(dt / 60 / 60 / 24, 1) % 30;
	    g_snprintf(
		buf, sizeof(buf),
		"%ld month%s %ld day%s",
		ct, ((ct == 1) ? "" : "s"),
		ct2, ((ct2 == 1) ? "" : "s")
	    );
	}     

	buf[sizeof(buf) - 1] = '\0';

	return(buf);
}


/*
 *	Prints the specified path and message to stderr.
 *
 *	This is intended to indicate and have the monitoring process
 *	record the object specified by path as having a virus or
 *	other problem specified by mesg.
 */
static void PRINT_ERR(const gchar *path, const gchar *mesg)
{
	g_printerr("\"%s\" %s\n", path, mesg);
}

/*
 *	Prints the specified path in a message to stdout to indicate
 *	it is being scanned.
 *
 *	This is intended to indicate and have the monitoring process
 *	display this message on its status.
 */
static void PRINT_SCANNING(const gchar *path)
{
	gchar *dpath = AVScanCopyShortenString(path, 60);
	g_print("Scanning: %s\n", dpath);
	g_free(dpath);
}
static void PRINT_SCANNING_DIR(const gchar *path)
{
	gchar *dpath = AVScanCopyShortenString(path, 50);
	g_print("Scanning directory: %s\n", dpath);
	g_free(dpath);
}

/*
 *	Prints the final report in a message to stdout.
 *
 *	This is intended to indicate and have the monitoring process
 *	display this message in a dialog.
 */
static void PRINT_REPORT(
	gulong time_start, gulong time_end,
	gulong infected_files,
	gulong size_scanned_mb,
	gulong files_scanned
)
{
	time_t t = (time_t)time_start;
	gchar	*s,
		*time_lapsed_s = STRDUP(TIME_LAPSED_STRING(
			time_end - time_start
		)),
		*time_start_s = STRDUP(ctime(&t)),
		*infected_files_s = STRDUP(FORMATED_NUMBER_STRING(
		    infected_files
		)),
		*size_scanned_mb_s = STRDUP(FORMATED_NUMBER_STRING(
		    size_scanned_mb
		)),
		*files_scanned_s = STRDUP(FORMATED_NUMBER_STRING(
		    files_scanned
		));

	/* Remove tailing newline character from time_start_s */
	for(s = time_start_s; *s != '\0'; s++)
	{
	    if(*s == '\n')
		*s = '\0';
	}

	/* Note that all ';' characters indicate newlines for the
	 * monitoring process that reads this
	 */
	if(infected_files > 0)
	    g_print(
"ReportInfected:\
Started On: %s;\
Duration: %s;\
;\
Infected Files: %s;\
Files Scanned: %s;\
Data Scanned: %s mb;\
;\
DO NOT VIEW OR RUN THE INFECTED FILES;\
\n",
		time_start_s,
		time_lapsed_s,
		infected_files_s,
		files_scanned_s,
		size_scanned_mb_s
	    );
	else
	    g_print(
"Report:\
Started On: %s;\
Duration: %s;\
;\
No infected files found;\
Files Scanned: %s;\
Data Scanned: %s mb;\
\n",
		time_start_s,
		time_lapsed_s,
		files_scanned_s,
		size_scanned_mb_s
	    );

	g_free(time_lapsed_s);
	g_free(time_start_s);
	g_free(infected_files_s);
	g_free(size_scanned_mb_s);
	g_free(files_scanned_s);
}

/*
 *	Returns a copy of the string that is no longer than max
 *	characters with appropriate shortened notation where 
 *	appropriate.
 */
static gchar *AVScanCopyShortenString(const gchar *s, gint max)
{
	gint len;

	if(s == NULL)  
	    return(NULL);

	len = STRLEN(s);
	if((len > max) && (max > 3))
	{
	    /* Need to shorten string */
	    gint i = len - max + 3;

	    return(g_strdup_printf(
		"...%s", (const gchar *)(s + i)
	    ));
	}      
	else
	{   
	    return(STRDUP(s));
	}
}


#if defined(CLAMAV_H)
/*
 *	Initializes the Clam AntiVirus.
 */
static ClamAVData *AVScanOPClamAVInit(const gchar *db_path)
{
	gint status;
	gchar *s;
	struct stat stat_buf;
	ClamAVData *d = (ClamAVData *)g_malloc0(
	    sizeof(ClamAVData)
	);

	d->files_scanned = 0l;
	d->size_scanned_mb = 0l;
	d->infected_files = 0l;

	/* Get virus database */
	if(db_path == NULL)
	    db_path = cl_retdbdir();
	if(db_path == NULL)
	{
	    g_print("Virus database location was not specified\n");
	    g_free(d);
	    return(NULL);
	}

	/* Load the virus database */ 
	s = AVScanCopyShortenString(db_path, 50);
	g_print("Loading virus database \"%s\"...\n", s);
	g_free(s);
	if((!stat(db_path, &stat_buf)) ? S_ISDIR(stat_buf.st_mode) : FALSE)
	    status = cl_loaddbdir(
		db_path, &d->database, &d->database_entries
	    );
	else
	    status = cl_loaddb(
		db_path, &d->database, &d->database_entries
	    );
	if(status || (d->database == NULL) || (d->database_entries <= 0))
	{
	    g_print("Failed to load the virus database\n");
	    g_free(d);
	    return(NULL);
	}

	/* Build the virus database */
 	if(cl_buildtrie(d->database))
	{
	    g_print("Failed to build the virus database\n");
	    g_free(d);
	    return(NULL);
	}

	return(d);
}

/*
 *	Shuts down the Clam AntiVirus.
 */
static void AVScanOPClamAVShutdown(ClamAVData *d)
{
	if(d == NULL)
	    return;

	/* Delete the virus database */
	if(d->database != NULL)
	    cl_freetrie(d->database);

	g_free(d);
}

/*
 *	Use Clam AntiVirus to scan the specified file.
 */
static gint AVScanOPClamAVScanFile(
	ClamAVData *d, const gchar *path,
	gint *stop_count
)
{
	gint status, fd;
	gchar *virus_name;
	gulong size_mb = 0l;
	struct cl_limits limits;

	PRINT_SCANNING(path);

	/* Open file for reading */
	fd = open(path, O_RDONLY);
	if(fd < 0)
	{
	    PRINT_ERR(path, "Unable to open file");
	    return(1);
	}

	/* Interrupted? */
	if(*stop_count > 0)
	{
	    close(fd);
	    return(4);
	}


	/* Set up archive limits */
	memset(&limits, 0x00, sizeof(struct cl_limits));
	limits.maxfiles = 1000;		/* Maximum files */
	limits.maxfilesize = 10 * 1048576;	/* Maximal archived file size == 10 Mb */
	limits.maxreclevel = 8;		/* Maximal recursion level */

	/* Scan this file */
	status = cl_scandesc(
	    fd, &virus_name, &size_mb,
	    d->database, &limits,
	    CL_ARCHIVE | CL_MAIL
	);

	/* Close file */
	close(fd);

	/* Count this file as scanned and how much scanned */
	d->files_scanned++;
	d->size_scanned_mb += size_mb;

	/* Check scan results */
	if(status == CL_VIRUS)
	{
	    /* Detected a virus */
	    gchar *buf = g_strdup_printf(
		"Detected virus \"%s\"",
		virus_name
	    );
	    PRINT_ERR(path, buf);
	    g_free(buf);

	    d->infected_files++;	/* Count this infected file */

	    return(1);
	}
	else if(status == 0)
	{
	    /* Clean */
	    return(0);
	}
	else
	{
	    /* Other error */
	    gchar *buf = g_strdup_printf(
		"%s", cl_perror(status)
	    );
	    PRINT_ERR(path, buf);
	    g_free(buf);
	    return(1);
	}
}

/*
 *	Use Clam AntiVirus to scan the specified directory.
 */
static gint AVScanOPClamAVScanDir(
	ClamAVData *d, const gchar *path,
	gboolean recursive,
	gboolean executables_only,
	gint *stop_count
)
{
	mode_t m;
	gint status, status2;
	const gchar *s;
	gchar *child;
	struct dirent *dent;
	struct stat stat_buf;
	DIR *dp = opendir(path);
	if(dp == NULL)
	{
	    PRINT_ERR(path, "Unable to scan directory");
	    return(1);
	}

	/* Scan directory */
	PRINT_SCANNING_DIR(path);
	status = 0;
	while(*stop_count == 0)
	{
	    dent = readdir(dp);
	    if(dent == NULL)
		break;

	    s = dent->d_name;

	    /* Skip special directory notations */
	    if(!g_strcasecmp(s, ".") ||
	       !g_strcasecmp(s, "..")
	    )
		continue;

	    /* Get full path to child object */
	    child = g_strdup_printf(
		"%s%c%s",
		path, DIR_DELIMINATOR, s
	    );

	    /* Get stats for this child object */
	    if(stat(child, &stat_buf))
	    {
		PRINT_ERR(child, "Unable to obtain stats");
		g_free(child);
		continue;
	    }

	    m = stat_buf.st_mode;

	    /* Is this child object a directory and should we recurse
	     * into it?
	     */
	    if(S_ISDIR(m) && recursive)
	    {
		status2 = AVScanOPClamAVScanDir(
		    d, child, recursive, executables_only,
		    stop_count
		);
		if((status2 != 0) && (status == 0))
		    status = status2;
	    }
	    else if(executables_only ?
		((m & S_IXUSR) || (m & S_IXGRP) || (m & S_IXOTH)) : TRUE
	    )
	    {
		status2 = AVScanOPClamAVScanFile(d, child, stop_count);
		if((status2 != 0) && (status == 0))
		    status = status2;
	    }

	    g_free(child);
	}

	closedir(dp);

	/* Iterrupted? */
	if((status == 0) && (*stop_count > 0))
	    status = 4;

	return(status);
}
#endif


/*
 *	AntiVirus Scan Operation
 *
 *	This is the front end to performing scans.
 */
gint AVScanOP(
	const gchar *db_path,
	GList *path_list,
	gboolean recursive,
	gboolean executables_only,
	gint *stop_count
)
{
	gint status = 0;
	gulong time_start = (gulong)time(NULL);
#if defined(CLAMAV_H)
	ClamAVData *d;
#endif

	if(path_list == NULL)
	{
	    status = 2;
	    return(status);
	}

	/* Interrupted? */
	if(*stop_count > 0)
	{
	    status = 4;
	    return(status);
	}

#if defined(CLAMAV_H)
	d = AVScanOPClamAVInit(db_path);
	if(d != NULL)
	{
	    gint status2;
	    const gchar *path;
	    GList *glist;
	    struct stat stat_buf;
	    for(glist = path_list; glist != NULL; glist = g_list_next(glist))
	    {
		path = (const gchar *)glist->data;
		if(!STRISEMPTY(path) ?
		    stat(path, &stat_buf) : TRUE
		)
		    continue;

		if(S_ISDIR(stat_buf.st_mode))
		    status2 = AVScanOPClamAVScanDir(
			d, path, recursive, executables_only,
			stop_count
		    );
		else
		    status2 = AVScanOPClamAVScanFile(
			d, path,
			stop_count
		    );
		if((status2 != 0) && (status == 0))
		    status = status2;
	    }
	    PRINT_REPORT(
		time_start,
		(gulong)time(NULL),
		d->infected_files,
		d->size_scanned_mb,
		d->files_scanned
	    );
	    AVScanOPClamAVShutdown(d);
	}
	else
	{
	    status = 3;
	}
#else
	g_print("No AntiVirus support compiled\n");
	status = 1;
#endif

	return(status);
}
